/************************************************************************
 *
 * SOS - Small Operating System
 *
 * This is an RTOS for TI MSP430 microcontrollers family.
 *
 * NOTICE: Copyright (C) 2002, 2003, 2004 by Oleg Skydan
 *
 * Files means any text, html, source or binary files that 
 * came with this copyright notice.
 *
 * You may use, embed, modify or distribute the contents
 * of the file(s) in any way you see fit as long as you
 * include this copyright notice in the resulting file(s).
 *
 * Any commercial or industrial use, except of SPECIAL EXCEPTION, 
 * including publishing or distributing SOS in any form IS 
 * PROHIBITED, and requires the written authorization of the author. 
 *
 * SPECIAL EXCEPTION:
 * If you choose to link SOS with other files to produce an
 * executable, this does not by itself cause the resulting 
 * executable to be covered by this license. Your use of the 
 * executable is in no way restricted on account of linking 
 * the SOS code with it. 
 *
 * SOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; WITH NO RESPONSIBILITY 
 * assumed by the author for bug fixes, support, or 
 * DAMAGES OF ANY SORT arising from the use of the SOS. 
 * 
 *
 * The program author can be contacted by E-mail at skidan@mail.ints.net
 *
 * For more details on the background or use of the program, please visit
 * http://users.ints.net/skidan/SOS
 *
 * OS kernel code
 ****************************************************************************/

#include "Kernel.h"


//Do we need to reschedule on exit from kernel mode?
uint8_t Reschedule;
//Active task
tid_t Running;
//Task control blocks
TCB Tasks[MAX_TASKS];

static uint8_t Quantum;

void __attribute__((naked)) __Enter(void)
{
#ifdef NO_IDLE_TASK
	asm volatile("tst.b %0           \n\t"\
               "jnz Save            \n\t"\
               "pop R10             \n\t"\
               "jmp NoSave          \n\t"\
          "Save:                    \n\t"\
               : : "m" ((uint8_t)Running));
#endif
   asm volatile("push R5	         	\n\t"\
               "push R6	         	\n\t"\
               "push R7	         	\n\t"\
               "push R8       		\n\t"\
               "push R9	         	\n\t"\
               "push R10          	\n\t"\
               "push R11      	   \n\t"\
               "push R13      	   \n\t"\
               "push R12      	   \n\t"\
               "push R15	         \n\t"\
               "push R14	         \n\t"\
               "mov 22(R1),R10	   \n\t"\
               "mov R4,22(R1)	      \n\t"\
               "mov.b %1,R11        \n\t"\
               "rla R11		         \n\t"\
               "rla R11		         \n\t"\
               "mov R1,Tasks(R11)	\n"\
        "NoSave:                    \n\t"\
               "mov #__stack,R1     \n\t"\
               "mov.b #0,%0	      \n\t"\
               "br R10			      \n\t"\
               : "=m" ((uint8_t)Reschedule)\
               : "m" ((uint8_t)Running));
}

void __attribute__((naked)) __Leave(void)
{
	if(Reschedule)
	{
		//We need to reschedule
		//Walk around the tasks begining from the next after active one
		//This will result in activation one of the highest priority tasks
		//(Preemptive sheduler) or if the current task has highest priority 
		//it will activate next task with the same priority (Round Robin sheduling)
		int t=0;
		for(int i=0,j=Running;i<MAX_TASKS-1;i++)
		{
			j++;
			if(j==MAX_TASKS)j=1;//We can skip Idle task.
								//It will activates if nothing else is ready.
			if(Tasks[j].state==TS_READY && Tasks[j].priority>Tasks[t].priority)t=j;
		}
		Running=t;
		Quantum=TASK_QUANTUM;
	}
#ifdef NO_IDLE_TASK
   if(Running==0)asm volatile("eint             \n\t"\
                              "br #__EnterLPM   \n\t" : :);
#endif
	asm volatile("mov.b &Running,R11    \n\t"\
				"rla R11		\n\t"\
				"rla R11		\n\t"\
				"mov Tasks(R11),R1 \n\t"\
				"pop R14		\n\t"\
				"pop R15		\n\t"\
				"pop R12		\n\t"\
				"pop R13		\n\t"\
				"pop R11		\n\t"\
				"pop R10		\n\t"\
				"pop R9		\n\t"\
				"pop R8		\n\t"\
				"pop R7		\n\t"\
				"pop R6		\n\t"\
				"pop R5		\n\t"\
				"pop R4		\n\t"\
				"reti			\n\t" : : );
}

#ifdef RR_SHEDULER

//Timer interrupt used for round robin sheduling
osinterrupt(TIMERA0_VECTOR) TimerA0(void)
{
	asm volatile("dec.b &Quantum	\n\t"\
				 "jz .LQuantumZero%=\n\t"\
				 "reti	\n\t"\
				 ".LQuantumZero%=:	\n\t" : :);
	ENTERINT();
	KReschedule();
	LEAVE();
}

#endif

tid_t KCreateTaskExt(void (*entry)(void),uint16_t* stack,uint8_t priority)
{
	for(int i=1;i<MAX_TASKS;i++)
	{
		if(Tasks[i].state==TS_EMPTY)
		{
			Tasks[i].priority=priority;
			Tasks[i].state=TS_READY;
			Tasks[i].stack=stack;
			*(--Tasks[i].stack)=(uint16_t)entry;
			*(--Tasks[i].stack)=0x08;//flags: GIE=1, others=0
			Tasks[i].stack-=12;
			KReschedule();
			return i;
		}
	}
	return K_ERR;
}

void KTerminateTask(tid_t tid)
{
	Tasks[tid].state=TS_EMPTY;
	KReschedule();
}

tid_t CreateTaskExt(void (*entry)(void),uint16_t* stack,uint8_t priority)
{
	ENTER();
   tid_t tid=KCreateTaskExt(entry,stack,priority);
	T_RES8(Running)=tid;
	LEAVE();
}

void TerminateTask(tid_t tid)
{
	ENTER();
	KTerminateTask(tid);
	LEAVE();
}

void KSuspendTask(tid_t tid)
{
	Tasks[tid].state|=TS_SUSPENDED;
	KReschedule();
}

void KResumeTask(tid_t tid)
{
	Tasks[tid].state&=~TS_SUSPENDED;
	KReschedule();
}

void SuspendTask(tid_t tid)
{
   ENTER();
   KSuspendTask(tid);
   LEAVE();
}

void ResumeTask(tid_t tid)
{
   ENTER();
   KResumeTask(tid);
   LEAVE();
}

void StartOS(void)
{
	Running=0;
	Tasks[0].state=TS_READY;

#ifndef NO_IDLE_TASK  //Prepare IDLE task if we use it
	Tasks[0].stack=Idle.stack;
	*(--Tasks[0].stack)=(uint16_t)Idle.entry;
	*(--Tasks[0].stack)=0x08;//flags: GIE=1, others=0
	Tasks[0].stack-=12;
#endif

	KReschedule();
	LEAVE();
}
