/************************************************************************
 *
 * SOS - Small Operating System
 *
 * This is an RTOS for TI MSP430 microcontrollers family.
 *
 * NOTICE: Copyright (C) 2002, 2003, 2004 by Oleg Skydan
 *
 * Files means any text, html, source or binary files that 
 * came with this copyright notice.
 *
 * You may use, embed, modify or distribute the contents
 * of the file(s) in any way you see fit as long as you
 * include this copyright notice in the resulting file(s).
 *
 * Any commercial or industrial use, except of SPECIAL EXCEPTION, 
 * including publishing or distributing SOS in any form IS 
 * PROHIBITED, and requires the written authorization of the author. 
 *
 * SPECIAL EXCEPTION:
 * If you choose to link SOS with other files to produce an
 * executable, this does not by itself cause the resulting 
 * executable to be covered by this license. Your use of the 
 * executable is in no way restricted on account of linking 
 * the SOS code with it. 
 *
 * SOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; WITH NO RESPONSIBILITY 
 * assumed by the author for bug fixes, support, or 
 * DAMAGES OF ANY SORT arising from the use of the SOS. 
 * 
 *
 * The program author can be contacted by E-mail at skidan@mail.ints.net
 *
 * For more details on the background or use of the program, please visit
 * http://users.ints.net/skidan/SOS
 *
 * Main OS header file
 ****************************************************************************/


#ifndef __OS_H_INCLUDED__
#define __OS_H_INCLUDED__

#include <sys/types.h>
#include "Config.h"


#define K_ERR 0xFF


typedef const struct
{
	void (*entry)(void);
	uint16_t* stack;
	uint8_t priority;
}TDB;


#define TS_SUSPENDED	0x01      //The task is suspended.

#define TS_PENDINGALL	0x02  //The task called WaitEvents() and is pending for all events
#define TS_PENDINGANY	0x04  //The task called WaitEvents() and is pending for any event(s)
#define TS_BLOCKED		0x08  //The task called EnterCS() and is pending for freeing 
							  //critical section by other task.
#define TS_READY		0x80  //Ready to go...
#define TS_EMPTY		0x00  //This is not task :), it's a free element in task list. 

typedef uint8_t tid_t;	//Task ID

//Macro for declaring tasks

#define DECLARE_TASK(name) extern TDB name

//Macro for Implement task
//Use:
//uint16_t stack[stack_size];
//TASK(name,stack,priority)
//{
//...task body...
//}

#define TASK(name,stack,priority) 	void __attribute__((naked)) entry##name(void); \
TDB name={entry##name,stack+sizeof(stack)/sizeof(stack[0]),priority};\
void entry##name(void)


#ifndef NO_IDLE_TASK

//Declare Idle task
DECLARE_TASK(Idle);
//Macro for implement Idle task
#define IDLE(stack) TASK(Idle,stack,0)

//Macro for implement Idle task
//Use:
//uint16_t stackIdle[stack_size];
//Idle(stackIdle)
//{
//...Idle task body...
//}

#else

//Define this function for entering idle mode
//NOTE: this function should never returns
void __attribute__((naked)) __EnterLPM(void);

#endif


//Start task kernel
void __attribute__((naked)) StartOS(void);

//Task functions
#define CreateTask(tdb) CreateTaskExt((tdb)->entry,(tdb)->stack,(tdb)->priority)
#define KCreateTask(tdb) KCreateTaskExt((tdb)->entry,(tdb)->stack,(tdb)->priority)
//Create new task
tid_t KCreateTaskExt(void (*entry)(void),uint16_t* stack,uint8_t priority);
//Create new task
tid_t __attribute__((naked)) CreateTaskExt(void (*entry)(void),uint16_t* stack,uint8_t priority);
//Get current task ID
#define GetTID() ({ extern tid_t Running; Running; })
//Set task priority
void __attribute__((naked)) SetTaskPriority(tid_t tid,uint8_t p);
//Get task priority
uint8_t GetTaskPriority(tid_t tid);
//Terminate task
void __attribute__((naked)) TerminateTask(tid_t tid);
//Suspend task execution
void __attribute__((naked)) SuspendTask(tid_t tid);
//Resume task execution
void __attribute__((naked)) ResumeTask(tid_t tid);

//Safe read long value
#define AtomicRead(x) ({ long _r;\
						 asm volatile("dint			\n\t" \
									  "nop			\n\t" \
									  "mov %A1,%A0	\n\t" \
									  "mov %B1,%B0	\n\t" \
									  "eint			\n\t" \
									  : "=r" ((long)_r) : "irm" ((long)x));\
									  _r;})

//Safe write long value
#define AtomicWrite(v,m) asm volatile("dint			\n\t" \
									  "nop			\n\t" \
									  "mov %A1,%A0	\n\t" \
									  "mov %B1,%B0	\n\t" \
									  "eint			\n\t" \
									  : "=m" ((long)m) : "irm" ((long)v))

//Safe addition for long value
#define AtomicAdd(v,m) asm volatile("dint			\n\t" \
									  "nop			 \n\t" \
									  "add  %A1,%A0 \n\t" \
									  "addc %B1,%B0 \n\t" \
									  "eint			 \n\t" \
									  : "=m" ((long)m) : "irm" ((long)v), "0" ((long)m))


// Enable/Disable interrupts do not use <signal.h> !!!
#define eint()  __asm__ __volatile__("eint"::)
#define dint()  __asm__ __volatile__("dint  \n\t"\
                                     "nop   \n\t"::)

//Interprocess communication functions

//Due to RAM economy I suggest that 15 normal and 1 idle task will be
//sufficient for msp430 project, so critical sections can be applyed only
//to first 16 task and may have 16 levels in depth.
 
typedef uint8_t cs_t; // ddddtttt - dddd-depth level,tttt-task number

void __attribute__((naked)) EnterCS(cs_t* cs);
void __attribute__((naked)) LeaveCS(cs_t* cs);

#define WAIT_ALL TS_PENDINGALL
#define WAIT_ANY TS_PENDINGANY

//Wait for single or multiple events (timers also), reset event(s) which are 
//auto reset events on exit
uint32_t  __attribute__((naked)) WaitEvents(uint32_t mask,uint8_t mode);
//Manual reset events
void critical ResetEvents(uint32_t mask);
//Check events
uint32_t critical TryEvents(uint32_t mask,uint8_t mode);
//Set events
void __attribute__((naked)) SetEvents(uint32_t mask);
//Set events type
void critical SetAutoEvents(uint32_t a);
void critical SetManualEvents(uint32_t m);



//Timer functions
//For perfomans reasons timer implemented in hardware (using TimerB block), rather then software.
//Advantages: faster,simpler,smaller :).
//Disadvantages: only 7 timers available for simultaneously use, TimerB block is occupied by
//timer.

typedef uint8_t timer_t;

#define ms(t) (t*TIMER_B_FRQ/1000)

//Timer types

#define TT_NOTIFY 0x01	//Notification timer event with manual reset
#define TT_SYNC	  0x02	//Syncronization timer auto reset event
#define TT_SUICIDAL  0x04  //Timer will kill itself after setting event
#define TT_PULSE  0x06  //Syncronization timer with automatic removal


void Sleep(uint16_t t);//Suspend task for t time period (for time in ms use ms() macro)
timer_t critical CreateTimer(uint8_t tt,uint16_t t,uint16_t cyc); //Create cyclic/simple timer
void critical SetTimer(timer_t timer,uint16_t t,uint16_t cyc); //Change timer timing
void KillTimer(timer_t t); //Kill timer

#endif
